use koopa::ir::{
    entities::ValueData,
    values::{Binary, Branch, Call, GlobalAlloc, Jump, Load, Return, Store},
    BinaryOp, FunctionData, Program, TypeKind, Value, ValueKind,
};

use crate::{
    asm_aux::{AsmAlloc, AsmFunction, AsmProgram, AsmValue, Emitter},
    Error, Result,
};

pub trait AsmGenerate<'a, 'p> {
    type AuxType;

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()>;
}

pub trait AsmValueGenerate<'a, 'p> {
    fn generate_asmvalue(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
    ) -> Result<AsmValue>;
}

impl<'a, 'p> AsmGenerate<'a, 'p> for Program {
    type AuxType = ();

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        // generate global allocations
        emitter.emit("  .data".into());
        dbg!(self.inst_layout());
        for &galloc in self.inst_layout() {
            let gvalue_data = self.borrow_value(galloc);
            let name = &gvalue_data.name().as_ref().unwrap()[1..];
            emitter.emit(format!("  .globl {}", name));
            emitter.emit(format!("{}:", name));
            gvalue_data.generate(emitter, asm_program, ())?;
            asm_program.insert_global_value(galloc, name.into());
        }

        emitter.emit("  .text".into());
        emitter.emit("  .globl main".into());

        for &func in self.func_layout() {
            let func_data = self.func(func);
            // skip function declaration
            if func_data.layout().entry_bb().is_none() {
                continue;
            }

            let asm_func = AsmFunction::new(func);

            asm_program.set_current_function(asm_func);
            func_data.generate(emitter, asm_program, ())?;
            let asm_func = asm_program.unset_current_function();
        }
        Ok(())
    }
}

impl<'a, 'p> AsmGenerate<'a, 'p> for FunctionData {
    type AuxType = ();

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        // Emit function name
        emitter.emit(format!("{}:", &self.name()[1..]));

        let cur_func_mut = asm_program.current_function_mut();
        // Scan the function and calculate stack space.
        let mut ra_size = 0;
        let mut max_param_num: isize = 0;
        for value_data in self.dfg().values().values() {
            if value_data.kind().is_local_inst() && !value_data.used_by().is_empty() {
                // we only consider instructions related with function stack.
                // and ignore value that's not used by any other values
                cur_func_mut.new_alloc(value_data);
            }

            match value_data.kind() {
                ValueKind::Call(call) => {
                    // if current function contains function call, then this is not a leaf function
                    ra_size = 4;

                    // max num of parameters
                    max_param_num = max_param_num.max(call.args().len() as isize);
                }
                _ => (),
            }
        }
        cur_func_mut.set_ra_size(ra_size);
        cur_func_mut.set_caller_stack_space_size(((max_param_num - 8).max(0) * 4) as usize);

        // Generate prologue
        cur_func_mut.init_func_sp_aligned_16();
        let function_stack_size = cur_func_mut.sp_offset();
        if function_stack_size > 2047 {
            // Load immediate number to temp register t0, and use add to update sp
            emitter.emit(format!("  li t0, {}", function_stack_size));
            emitter.emit(format!("  sub sp, sp, t0"));
            // Then t0 is allowed to be polluted
        } else {
            emitter.emit(format!(
                "  addi sp, sp, {}",
                -(function_stack_size as isize)
            ));
        }

        // generate ra store if not leaf
        if !cur_func_mut.is_leaf() {
            let ra_offset = cur_func_mut.ra_offset_relative_to_sp();
            if ra_offset > 2047 {
                emitter.emit(format!("  li t0, {}", ra_offset));
                emitter.emit(format!("  add t0, t0, sp"));
                emitter.emit(format!("  sw ra, 0(t0)"));
            } else {
                emitter.emit(format!("  sw ra, {}(sp)", ra_offset));
            }
        }

        for (i, (&bb, _)) in self.layout().bbs().iter().enumerate() {
            let bb_data = self.dfg().bbs().get(&bb).unwrap();
            cur_func_mut.push_bb(bb, format!("{}_{}", bb_data.name().as_ref().unwrap(), i));
        }

        for (&bb, node) in self.layout().bbs() {
            let Some(name) = asm_program.current_function().bb_name(bb) else {
                return Err(Error::Asm(
                    "Jump".into(),
                    "Do not find basic block in current function".into(),
                ));
            };
            emitter.emit(format!("{}:", &name[1..]));
            for &inst in node.insts().keys() {
                let value_data = self.dfg().value(inst);
                value_data.generate(emitter, asm_program, ())?;
            }
        }

        Ok(())
    }
}

// Give value
impl<'a, 'p> AsmValueGenerate<'a, 'p> for Value {
    fn generate_asmvalue(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
    ) -> Result<AsmValue> {
        if self.is_global() {
            let global_value_name = asm_program.get_global_value_name(*self);
            let res = AsmValue::Global(global_value_name);
            return Ok(res);
        }

        let cur_func = asm_program.current_function();
        let value_data = asm_program
            .program()
            .func(cur_func.func())
            .dfg()
            .value(*self);
        let res = match value_data.kind() {
            ValueKind::Integer(integer) => AsmValue::Integer(integer.value()), // e.g. store 0, %ret: 0
            ValueKind::FuncArgRef(func_arg_ref) => AsmValue::FuncArg(func_arg_ref.index()),
            _ => match cur_func.value_allocated(value_data) {
                // e.g. store 0, %ret: %ret
                // No allocated value in current function, then just return a void
                None => AsmValue::Void,
                // Found allocated value in current function (e.g. %ret), return the local
                Some(asm_alloc) => AsmValue::Local(asm_alloc),
            },
        };
        Ok(res)
    }
}

// Emit instruction
impl<'a, 'p> AsmGenerate<'a, 'p> for ValueData {
    type AuxType = ();

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        match self.kind() {
            ValueKind::Aggregate(aggregate) => todo!(),
            ValueKind::Alloc(alloc) => return Ok(()), // already processed in function prologue stage
            ValueKind::Binary(binary) => binary.generate(emitter, asm_program, self)?,
            ValueKind::BlockArgRef(block_arg_ref) => todo!(),
            ValueKind::Branch(branch) => branch.generate(emitter, asm_program, ())?,
            ValueKind::Call(call) => call.generate(emitter, asm_program, self)?,
            ValueKind::FuncArgRef(func_arg_ref) => todo!(),
            ValueKind::GetElemPtr(get_elem_ptr) => todo!(),
            ValueKind::GetPtr(get_ptr) => todo!(),
            ValueKind::GlobalAlloc(global_alloc) => {
                global_alloc.generate(emitter, asm_program, ())?
            }
            ValueKind::Integer(integer) => {
                emitter.emit(format!("  .word {}", integer.value()));
            }
            ValueKind::Jump(jump) => jump.generate(emitter, asm_program, ())?,
            ValueKind::Load(load) => load.generate(emitter, asm_program, self)?,
            ValueKind::Return(ret) => ret.generate(emitter, asm_program, ())?,
            ValueKind::Store(store) => store.generate(emitter, asm_program, ())?,
            ValueKind::Undef(undef) => todo!(),
            ValueKind::ZeroInit(zero_init) => {
                emitter.emit(format!("  .zero {}", self.ty().size()));
            }
        }
        Ok(())
    }
}

impl<'a, 'p> AsmGenerate<'a, 'p> for GlobalAlloc {
    type AuxType = ();
    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        let value = asm_program.program().borrow_value(self.init()).clone();
        value.generate(emitter, asm_program, ())?;
        Ok(())
    }
}

// Store do not have return value
// sw rd, offset(rs1)
impl<'a, 'p> AsmGenerate<'a, 'p> for Store {
    type AuxType = ();

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        let v = self.value().generate_asmvalue(emitter, asm_program)?;
        let dst = self.dest().generate_asmvalue(emitter, asm_program)?;

        let cur_func = asm_program.current_function();

        match v {
            AsmValue::Void => {
                return Err(Error::Asm("Store".into(), "storing void to memory".into()))
            }
            AsmValue::Integer(integer) => {
                // When `sw rd, offset(rs1)`, value must inside a register, if not then should `li` it to t0
                emitter.emit(format!("  li t0, {}", integer));
                // emitter.emit(format!("  sw t0, {}(sp)"));
            }
            AsmValue::Local(asm_alloc) => {
                // Currently there's no register allocation algorithm.
                // All the local asm values must be loaded from stack into t0, and then store to destination.
                // Constant propagation optimization needed!
                let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                if sp_offset > 2047 {
                    emitter.emit(format!("  li t0, {}", sp_offset));
                    emitter.emit(format!("  mv t1, sp"));
                    emitter.emit(format!("  add t1, t0, t1"));
                    emitter.emit(format!("  lw t0, 0(t1)"));
                } else {
                    emitter.emit(format!("  lw t0, {}(sp)", sp_offset));
                }
            }
            AsmValue::Global(global_value_name) => {
                emitter.emit(format!("  la t1, {}", global_value_name));
                emitter.emit(format!("  lw t0, 0(t1)"));
            }
            AsmValue::FuncArg(idx) => {
                // the value to be moved out of function argument
                if idx < 8 {
                    // in register
                    emitter.emit(format!("  mv t0, a{}", idx));
                } else {
                    // in stack
                    let arg_offset = cur_func.arg_from_caller_offset_to_sp(idx);
                    if arg_offset > 2047 {
                        emitter.emit(format!("  li t0, {}", arg_offset));
                        emitter.emit(format!("  mv t1, sp"));
                        emitter.emit(format!("  add t1, t0, t1"));
                        emitter.emit(format!("  lw t0, 0(t1)"));
                    } else {
                        emitter.emit(format!("  lw t0, {}(sp)", arg_offset));
                    }
                }
            }
        }

        // sw rd, offset(rs1)
        match dst {
            // We consider this is as an error
            AsmValue::Void => {
                return Err(Error::Asm(
                    "Store".into(),
                    "store destination is void!".into(),
                ))
            }
            // How could this be?
            AsmValue::Integer(_integer) => {
                return Err(Error::Asm(
                    "Store".into(),
                    "store destination is an integer!".into(),
                ))
            }
            AsmValue::Local(asm_alloc) => {
                // we've already have the value to be stored loaded inside t0
                // t0 already occupied!
                let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                if sp_offset > 2047 {
                    emitter.emit(format!("  li t2, {}", sp_offset));
                    emitter.emit(format!("  mv t3, sp"));
                    emitter.emit(format!("  add t3, t2, t3"));
                    emitter.emit(format!("  sw t0, 0(t3)"));
                } else {
                    emitter.emit(format!("  sw t0, {}(sp)", sp_offset));
                }
            }
            AsmValue::Global(global_value_name) => {
                emitter.emit(format!("  la t1, {}", global_value_name));
                emitter.emit(format!("  sw t0, 0(t1)"));
            }
            AsmValue::FuncArg(_) => {
                return Err(Error::Asm(
                    "Store".into(),
                    "Storing to a function argument?".into(),
                ))
            }
        }

        Ok(())
    }
}

// Load have a return value
// lw rd, offset(rs1)
impl<'a, 'p> AsmGenerate<'a, 'p> for Load {
    type AuxType = &'a ValueData;

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        let src = self.src().generate_asmvalue(emitter, asm_program)?;
        let cur_func = asm_program.current_function();

        // Check the return value of Load
        // It shall be a AsmValue::Local
        let load_ret = cur_func.value_allocated(aux);
        // if no destination
        if load_ret.is_none() {
            return Ok(());
        }

        // Load source shall be a AsmValue::Local
        match src {
            AsmValue::Local(asm_alloc) => {
                let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                if sp_offset > 2047 {
                    // Load immediate number to temp register t0
                    // Load sp register to temp register t1
                    // Calculate the memory destination

                    // Probability that need to save t0 and t1 first?
                    emitter.emit(format!("  li t0, {}", sp_offset));
                    emitter.emit(format!("  mv t1, sp"));
                    emitter.emit(format!("  add t1, t0, t1"));
                    emitter.emit(format!("  lw t0, 0(t1)"));
                } else {
                    emitter.emit(format!("  lw t0, {}(sp)", sp_offset));
                }
            }
            AsmValue::Global(global_value_name) => {
                // la t0, var
                // lw t0, 0(t0)
                emitter.emit(format!("  la t1, {}", global_value_name));
                emitter.emit(format!("  lw t0, 0(t1)"));
            }
            AsmValue::FuncArg(_) => unimplemented!(),
            _ => {
                return Err(Error::Asm(
                    "Load".into(),
                    "Load source shall be a AsmValue::Local or AsmValue::Global".into(),
                ))
            }
        }

        match load_ret {
            None => {
                // Just do nothing
                /*
                int main() {
                    1 || 2;
                    return 0;
                }
                */
                // return Err(Error::Asm(
                //     "Load".into(),
                //     "Load ret not in function allocations".into(),
                // ))
            }
            Some(asm_alloc) => {
                // Here we should store the loaded value into stack again.
                // Need optimization!
                let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                if sp_offset > 2047 {
                    // Load immediate number to temp register t0
                    // Load sp register to temp register t1
                    // Calculate the memory destination

                    // Probability that need to save t0 and t1 first?
                    emitter.emit(format!("  li t2, {}", sp_offset));
                    emitter.emit(format!("  mv t3, sp"));
                    emitter.emit(format!("  add t3, t2, t3"));
                    emitter.emit(format!("  sw t0, 0(t3)"));
                } else {
                    emitter.emit(format!("  sw t0, {}(sp)", sp_offset));
                }
            }
        }

        Ok(())
    }
}

impl<'a, 'p> AsmGenerate<'a, 'p> for Return {
    type AuxType = ();

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        // let Some(ret_val) = self.value() else {
        //     emitter.emit(format!("  ret"));
        //     return Ok(());
        // };

        if let Some(ret_val) = self.value() {
            let ret_val = ret_val.generate_asmvalue(emitter, asm_program)?;
            let cur_func = asm_program.current_function();

            match ret_val {
                AsmValue::Void => {
                    return Err(Error::Asm(
                        "Return".into(),
                        "Return value is not None but get AsmValue::Void".into(),
                    ))
                }
                AsmValue::Integer(integer) => {
                    emitter.emit(format!("  li a0, {}", integer));
                }
                AsmValue::Local(asm_alloc) => {
                    // Need constant propagation!
                    let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                    // emitter.emit(format!("  lw a0, {}(sp)", sp_offset));

                    if sp_offset > 2047 {
                        emitter.emit(format!("  li t2, {}", sp_offset));
                        emitter.emit(format!("  mv t3, sp"));
                        emitter.emit(format!("  add t3, t2, t3"));
                        emitter.emit(format!("  lw a0, 0(t3)"));
                    } else {
                        emitter.emit(format!("  lw a0, {}(sp)", sp_offset));
                    }
                }
                AsmValue::Global(global_value_name) => {
                    emitter.emit(format!("  la t1, {}", global_value_name));
                    emitter.emit(format!("  lw a0, 0(t1)"));
                }
                AsmValue::FuncArg(_) => unimplemented!(),
            }
        }

        // Generate epilogue
        // generate ra store if not leaf
        let cur_func = asm_program.current_function();

        if !cur_func.is_leaf() {
            let ra_offset = cur_func.ra_offset_relative_to_sp();
            if ra_offset > 2047 {
                emitter.emit(format!("  li t0, {}", ra_offset));
                emitter.emit(format!("  add t0, t0, sp"));
                emitter.emit(format!("  lw ra, 0(t0)"));
            } else {
                emitter.emit(format!("  lw ra, {}(sp)", ra_offset));
            }
        }

        let function_stack_size = cur_func.sp_offset();
        if function_stack_size > 2047 {
            // Load immediate number to temp register t0, and use add to update sp
            emitter.emit(format!("  li t0, {}", function_stack_size));
            emitter.emit(format!("  add sp, sp, t0"));
            // Then t0 is allowed to be polluted
        } else {
            emitter.emit(format!("  addi sp, sp, {}", function_stack_size));
        }

        emitter.emit(format!("  ret"));

        Ok(())
    }
}

impl<'a, 'p> AsmGenerate<'a, 'p> for Jump {
    type AuxType = ();

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        let jump_label = self.target();
        let Some(name) = asm_program.current_function().bb_name(jump_label) else {
            return Err(Error::Asm(
                "Jump".into(),
                "Do not find basic block in current function".into(),
            ));
        };

        emitter.emit(format!("  j {}", &name[1..]));
        Ok(())
    }
}

impl<'a, 'p> AsmGenerate<'a, 'p> for Binary {
    type AuxType = &'a ValueData;

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        let lhs = self.lhs().generate_asmvalue(emitter, asm_program)?;
        let rhs = self.rhs().generate_asmvalue(emitter, asm_program)?;
        let cur_func = asm_program.current_function();
        match lhs {
            AsmValue::Void => panic!(),
            AsmValue::Integer(integer) => {
                emitter.emit(format!("  li t0, {}", integer));
            }
            AsmValue::Local(asm_alloc) => {
                let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                if sp_offset > 2047 {
                    emitter.emit(format!("  li t0, {}", sp_offset));
                    emitter.emit(format!("  mv t1, sp"));
                    emitter.emit(format!("  add t1, t0, t1"));
                    emitter.emit(format!("  lw t0, 0(t1)"));
                } else {
                    emitter.emit(format!("  lw t0, {}(sp)", sp_offset));
                }
            }
            AsmValue::Global(global_value_name) => {
                emitter.emit(format!("  la t1, {}", global_value_name));
                emitter.emit(format!("  lw t0, 0(t1)"));
            }
            AsmValue::FuncArg(_) => unimplemented!(),
        }

        match rhs {
            AsmValue::Void => panic!(),
            AsmValue::Integer(integer) => {
                emitter.emit(format!("  li t1, {}", integer));
            }
            AsmValue::Local(asm_alloc) => {
                let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                if sp_offset > 2047 {
                    emitter.emit(format!("  li t2, {}", sp_offset));
                    emitter.emit(format!("  mv t3, sp"));
                    emitter.emit(format!("  add t3, t2, t3"));
                    emitter.emit(format!("  lw t1, 0(t3)"));
                } else {
                    emitter.emit(format!("  lw t1, {}(sp)", sp_offset));
                }
            }
            AsmValue::Global(global_value_name) => {
                emitter.emit(format!("  la t2, {}", global_value_name));
                emitter.emit(format!("  lw t1, 0(t2)"));
            }
            AsmValue::FuncArg(_) => unimplemented!(),
        }

        match self.op() {
            BinaryOp::Add => {
                emitter.emit(format!("  add t0, t0, t1"));
            }
            BinaryOp::And => {
                emitter.emit(format!("  and t0, t0, t1"));
            }
            BinaryOp::Div => {
                emitter.emit(format!("  div t0, t0, t1"));
            }
            BinaryOp::Eq => {
                emitter.emit(format!("  xor t0, t0, t1"));
                emitter.emit(format!("  seqz t0, t0"));
            }
            BinaryOp::Ge => {
                emitter.emit(format!("  slt t0, t0, t1"));
                emitter.emit(format!("  seqz t0, t0"));
            }
            BinaryOp::Gt => {
                emitter.emit(format!("  sgt t0, t0, t1"));
            }
            BinaryOp::Le => {
                emitter.emit(format!("  sgt t0, t0, t1"));
                emitter.emit(format!("  seqz t0, t0"));
            }
            BinaryOp::Lt => {
                emitter.emit(format!("  slt t0, t0, t1"));
            }
            BinaryOp::Mod => {
                emitter.emit(format!("  rem t0, t0, t1"));
            }
            BinaryOp::Mul => {
                emitter.emit(format!("  mul t0, t0, t1"));
            }
            BinaryOp::NotEq => {
                emitter.emit(format!("  xor t0, t0, t1"));
                emitter.emit(format!("  snez t0, t0"));
            }
            BinaryOp::Or => {
                emitter.emit(format!("  or t0, t0, t1"));
            }
            BinaryOp::Sar => {
                emitter.emit(format!("  sra t0, t0, t1"));
            }
            BinaryOp::Shl => {
                emitter.emit(format!("  sll t0, t0, t1"));
            }
            BinaryOp::Shr => {
                emitter.emit(format!("  srl t0, t0, t1"));
            }
            BinaryOp::Sub => {
                emitter.emit(format!("  sub t0, t0, t1"));
            }
            BinaryOp::Xor => {
                emitter.emit(format!("  xor t0, t0, t1"));
            }
        }

        let binary_ret = cur_func.value_allocated(aux);
        match binary_ret {
            None => {
                // Expression result discarded
                // Just do nothing
            }
            Some(asm_alloc) => {
                // Here we should store the loaded value into stack again.
                // Need optimization!

                // t0 already occupied!
                let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                if sp_offset > 2047 {
                    emitter.emit(format!("  li t2, {}", sp_offset));
                    emitter.emit(format!("  mv t3, sp"));
                    emitter.emit(format!("  add t3, t2, t3"));
                    emitter.emit(format!("  sw t0, 0(t3)"));
                } else {
                    emitter.emit(format!("  sw t0, {}(sp)", sp_offset));
                }
            }
        }
        Ok(())
    }
}

impl<'a, 'p> AsmGenerate<'a, 'p> for Branch {
    type AuxType = ();

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        // Load cond into t0
        let cond = self.cond().generate_asmvalue(emitter, asm_program)?;
        let cur_func = asm_program.current_function();
        match cond {
            AsmValue::Void => {
                return Err(Error::Asm(
                    "Branch".into(),
                    "Condition is AsmValue::Void".into(),
                ))
            }
            AsmValue::Integer(integer) => {
                emitter.emit(format!("  li t0, {}", integer));
            }
            AsmValue::Local(asm_alloc) => {
                let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                if sp_offset > 2047 {
                    emitter.emit(format!("  li t0, {}", sp_offset));
                    emitter.emit(format!("  mv t1, sp"));
                    emitter.emit(format!("  add t1, t0, t1"));
                    emitter.emit(format!("  lw t0, 0(t1)"));
                } else {
                    emitter.emit(format!("  lw t0, {}(sp)", sp_offset));
                }
            }
            AsmValue::Global(global_value_name) => {
                emitter.emit(format!("  la t1, {}", global_value_name));
                emitter.emit(format!("  lw t0, 0(t1)"));
            }
            AsmValue::FuncArg(_) => unimplemented!(),
        }

        let then_name = cur_func.bb_name(self.true_bb()).unwrap();
        let else_name = cur_func.bb_name(self.false_bb()).unwrap();
        emitter.emit(format!("  bnez t0, {}", &then_name[1..]));
        emitter.emit(format!("  j {}", &else_name[1..]));

        Ok(())
    }
}

impl<'a, 'p> AsmGenerate<'a, 'p> for Call {
    type AuxType = &'a ValueData;

    fn generate(
        &self,
        emitter: &impl Emitter,
        asm_program: &'a mut AsmProgram<'p>,
        aux: Self::AuxType,
    ) -> Result<()> {
        // TODO: store caller saved registers

        // generate parameters
        let mut args = vec![];
        for v in self.args().iter() {
            let arg = v.generate_asmvalue(emitter, asm_program)?;
            args.push(arg);
        }

        let cur_func = asm_program.current_function();

        // prepare arguments
        let register_args_len = args.len().min(8);
        for idx in 0..register_args_len {
            match &args[idx] {
                AsmValue::Local(asm_alloc) => {
                    let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                    if sp_offset > 2047 {
                        emitter.emit(format!("  li t0, {}", sp_offset));
                        emitter.emit(format!("  mv t1, sp"));
                        emitter.emit(format!("  add t1, t0, t1"));
                        emitter.emit(format!("  lw a{}, 0(t1)", idx));
                    } else {
                        emitter.emit(format!("  lw a{}, {}(sp)", idx, sp_offset));
                    }
                }
                AsmValue::Integer(integer) => {
                    emitter.emit(format!("  li a{}, {}", idx, integer));
                }
                AsmValue::Void => {
                    return Err(Error::Asm(
                        "Call".into(),
                        "function argument type is void".into(),
                    ))
                }
                AsmValue::Global(global_value_name) => {
                    emitter.emit(format!("  la t1, {}", global_value_name));
                    emitter.emit(format!("  lw a{}, 0(t1)", idx));
                }
                AsmValue::FuncArg(_) => panic!(),
            }
        }

        if args.len() > 8 {
            for idx in 8..args.len() {
                match &args[idx] {
                    AsmValue::Local(asm_alloc) => {
                        let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                        if sp_offset > 2047 {
                            emitter.emit(format!("  li t0, {}", sp_offset));
                            emitter.emit(format!("  mv t1, sp"));
                            emitter.emit(format!("  add t1, t0, t1"));
                            emitter.emit(format!("  lw t0, 0(t1)"));
                        } else {
                            emitter.emit(format!("  lw t0, {}(sp)", sp_offset));
                        }

                        // already in t0
                        let arg_offset = cur_func.arg_prepared_for_callee_offset_to_sp(idx);
                        if arg_offset > 2047 {
                            emitter.emit(format!("  li t2, {}", arg_offset));
                            emitter.emit(format!("  mv t3, sp"));
                            emitter.emit(format!("  add t3, t2, t3"));
                            emitter.emit(format!("  sw t0, 0(t3)"));
                        } else {
                            emitter.emit(format!("  sw t0, {}(sp)", arg_offset));
                        }
                    }
                    AsmValue::Integer(integer) => {
                        emitter.emit(format!("  li t0, {}", integer));
                        emitter.emit(format!(
                            "  sw t0, {}(sp)",
                            cur_func.arg_prepared_for_callee_offset_to_sp(idx)
                        ));
                    }
                    AsmValue::Void => {
                        return Err(Error::Asm(
                            "Call".into(),
                            "function argument type is void".into(),
                        ))
                    }
                    AsmValue::Global(global_value_name) => {
                        emitter.emit(format!("  la t1, {}", global_value_name));
                        emitter.emit(format!("  lw t0, 0(t1)"));

                        // already in t0
                        let arg_offset = cur_func.arg_prepared_for_callee_offset_to_sp(idx);
                        if arg_offset > 2047 {
                            emitter.emit(format!("  li t2, {}", arg_offset));
                            emitter.emit(format!("  mv t3, sp"));
                            emitter.emit(format!("  add t3, t2, t3"));
                            emitter.emit(format!("  sw t0, 0(t3)"));
                        } else {
                            emitter.emit(format!("  sw t0, {}(sp)", arg_offset));
                        }
                    }
                    AsmValue::FuncArg(_) => panic!(),
                }
            }
        }

        // call
        let target_func = asm_program.program().func(self.callee());
        emitter.emit(format!("  call {}", &target_func.name()[1..]));

        // handle return value
        if aux.used_by().is_empty() {
            // no one care about the return value
        } else {
            // check the return value and search it in current function context
            if let Some(asm_alloc) = cur_func.value_allocated(aux) {
                let sp_offset = cur_func.offset_relative_to_sp(&asm_alloc);
                if sp_offset > 2047 {
                    emitter.emit(format!("  li t0, {}", sp_offset));
                    emitter.emit(format!("  mv t1, sp"));
                    emitter.emit(format!("  add t1, t0, t1"));
                    emitter.emit(format!("  sw a0, 0(t1)"));
                } else {
                    emitter.emit(format!("  sw a0, {}(sp)", sp_offset));
                }
            }
        }

        Ok(())
    }
}
