use koopa::ir::{
    entities::ValueData, BasicBlock, Function, FunctionData, Program, Value, ValueKind,
};
use std::{cell::RefCell, collections::HashMap, fs, io::Write};

pub trait Emitter {
    fn emit(&self, line: String);
}

pub struct StdoutEmitter {}

impl StdoutEmitter {
    pub fn new() -> Self {
        Self {}
    }
}

impl Emitter for StdoutEmitter {
    fn emit(&self, line: String) {
        println!("{}", line);
    }
}

pub struct FileEmitter {
    file: RefCell<fs::File>,
}

impl FileEmitter {
    pub fn new(file: fs::File) -> Self {
        Self {
            file: RefCell::new(file),
        }
    }
}

impl Emitter for FileEmitter {
    fn emit(&self, line: String) {
        let mut inner = self.file.borrow_mut();
        inner.write_all(line.as_bytes()).unwrap();
        inner.write_all(b"\n").unwrap();
    }
}

pub struct AsmProgram<'p> {
    program: &'p Program,
    current_function: Option<AsmFunction>,
    functions: HashMap<*const FunctionData, AsmFunction>,
    global_values: HashMap<Value, String>,
}

impl<'p> AsmProgram<'p> {
    pub fn new(program: &'p Program) -> Self {
        Self {
            program,
            current_function: None,
            functions: HashMap::new(),
            global_values: HashMap::new(),
        }
    }

    pub fn program(&self) -> &Program {
        self.program
    }

    pub fn set_current_function(&mut self, f: AsmFunction) {
        self.current_function = Some(f);
    }

    pub fn unset_current_function(&mut self) -> AsmFunction {
        self.current_function.take().unwrap()
    }

    pub fn current_function(&self) -> &AsmFunction {
        self.current_function.as_ref().unwrap()
    }

    pub fn current_function_mut(&mut self) -> &mut AsmFunction {
        self.current_function.as_mut().unwrap()
    }

    pub fn get_function(&self, func_data: &FunctionData) -> Option<&AsmFunction> {
        self.functions.get(&(func_data as *const _))
    }

    pub fn insert_global_value(&mut self, value: Value, name: String) {
        self.global_values.insert(value, name);
    }

    pub fn get_global_value_name(&self, value: Value) -> String {
        self.global_values
            .get(&value)
            .cloned()
            .expect(&format!("No such value: {:?}", value))
    }
}

#[derive(Debug)]
pub struct AsmFunction {
    func: Function,
    ra_size: usize,
    caller_stack_space_size: usize,
    sp_offset: usize,                 // sp offset of current function
    bbs: HashMap<BasicBlock, String>, // for jumps
    allocation_size: usize,           // size that's need to be allocated within the function
    allocations: HashMap<*const ValueData, AsmAlloc>, // ValueData - offset in the stack
}

#[derive(Debug, Clone, Copy)]
pub struct AsmAlloc {
    alloc_size: usize, // how many bytes does this allocation occupies
    offset: usize,
}

impl AsmFunction {
    pub fn new(func: Function) -> Self {
        Self {
            func,
            ra_size: 4,
            caller_stack_space_size: 0,
            sp_offset: 0,
            bbs: HashMap::new(),
            allocation_size: 0,
            allocations: HashMap::new(),
        }
    }

    pub fn is_leaf(&self) -> bool {
        self.ra_size == 0
    }

    pub fn ra_size(&self) -> usize {
        self.ra_size
    }

    pub fn set_ra_size(&mut self, ra_size: usize) {
        self.ra_size = ra_size;
    }

    pub fn caller_stack_space_size(&self) -> usize {
        self.caller_stack_space_size
    }

    pub fn set_caller_stack_space_size(&mut self, s: usize) {
        self.caller_stack_space_size = s;
    }

    pub fn func(&self) -> Function {
        self.func
    }

    pub fn sp_offset(&self) -> usize {
        self.sp_offset
    }

    pub fn init_func_sp_aligned_16(&mut self) {
        // self.sp_offset = self.allocation_size.next_multiple_of(16) // rust 1.70.0 do not support
        self.sp_offset =
            ((self.ra_size + self.caller_stack_space_size + self.allocation_size + 15) / 16) * 16;
    }

    // We choose to align to stack top
    pub fn offset_relative_to_sp(&self, asm_alloc: &AsmAlloc) -> usize {
        asm_alloc.offset + self.caller_stack_space_size
    }

    // Get the position of `n`th argument prepared for callee, where n >= 8
    pub fn arg_prepared_for_callee_offset_to_sp(&self, arg_index: usize) -> usize {
        assert!(arg_index >= 8);
        (arg_index - 8) * 4
    }

    // Get the position of `n`th argument prepared by caller, where n >= 8
    pub fn arg_from_caller_offset_to_sp(&self, arg_index: usize) -> usize {
        assert!(arg_index >= 8);
        (arg_index - 8) * 4 + self.sp_offset
    }

    pub fn ra_offset_relative_to_sp(&self) -> usize {
        assert!(!self.is_leaf());
        self.sp_offset - 4
    }

    pub fn push_bb(&mut self, bb: BasicBlock, name: String) {
        self.bbs.insert(bb, name);
    }

    pub fn new_alloc(&mut self, value_data: &ValueData) {
        // self.allocations.insert(value_data, offset);

        let alloc_size = match value_data.kind() {
            // 1. This is allocation
            ValueKind::Alloc(_alloc) => {
                // currently only consider integer, so 4 bytes
                let alloc_size = 4;

                self.allocations.insert(
                    value_data as *const _,
                    AsmAlloc {
                        alloc_size,
                        offset: self.allocation_size,
                    },
                );
                alloc_size
            }
            // 2. This is instruction with return value
            // e.g.
            // %0 = load @x
            // %1 = add %0, 1
            _ => {
                if value_data.ty().is_unit() {
                    return; // do not allocate
                }
                let alloc_size = 4;
                self.allocations.insert(
                    value_data as *const _,
                    AsmAlloc {
                        alloc_size,
                        offset: self.allocation_size,
                    },
                );

                alloc_size
            }
        };

        self.allocation_size += alloc_size;
    }

    pub fn value_allocated(&self, value_data: &ValueData) -> Option<AsmAlloc> {
        self.allocations.get(&(value_data as *const _)).cloned()
    }

    pub fn bb_name(&self, bb: BasicBlock) -> Option<&String> {
        self.bbs.get(&bb)
    }

    pub fn print(&self) {
        dbg!(self);
    }
}

pub enum AsmValue {
    Global(String),  // Global identifier
    FuncArg(usize),  // FuncArg with index
    Integer(i32),    // Integer
    Local(AsmAlloc), // Local stack (offset)
    Void,            // Void
}
