//! Constantly evaluate values

use koopa::ir::{Function, Value};

use crate::{
    ast::{
        AddExp, ConstExp, ConstInitVal, EqExp, Exp, InitVal, LAndExp, LOrExp, LVal, MulExp, PrimaryExp, RelExp, UnaryExp, UnaryOp
    },
    scopes::{IRFunction, Scopes},
};

#[derive(Debug, Clone)]
pub enum ExpValue {
    Undefined,
    InitVal(Value),
    Integer(i32),
    Function(ExpFunction),
}

#[derive(Debug, Clone)]
pub struct ExpFunction {
    name: String,
    func: Function,

}

impl ExpFunction {
    pub fn new(name: String, func: Function) -> Self {
        Self { name, func }
    }

    pub fn func(&self) -> Function {
        self.func
    }

    pub fn name(&self) -> &String {
        &self.name
    }
}

pub trait ExpEval {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32>;
}

impl ExpEval for InitVal {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            InitVal::Exp(exp) => exp.eval(scopes),
            InitVal::Array(array) => todo!(),
        }
    }
}

impl ExpEval for ConstInitVal {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            ConstInitVal::Exp(const_exp) => const_exp.eval(scopes),
            ConstInitVal::Array(array) => todo!(),
        }
    }
}

impl ExpEval for ConstExp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            ConstExp::Exp(exp) => exp.eval(scopes),
        }
    }
}

impl ExpEval for AddExp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            AddExp::Add(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some(lhs + rhs),
                    _ => None,
                }
            }
            AddExp::MulExp(mul_exp) => mul_exp.eval(scopes),
            AddExp::Sub(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some(lhs - rhs),
                    _ => None,
                }
            }
        }
    }
}

impl ExpEval for MulExp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            MulExp::Div(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some(lhs / rhs),
                    _ => None,
                }
            }
            MulExp::Mod(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some(lhs % rhs),
                    _ => None,
                }
            }
            MulExp::Mul(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some(lhs * rhs),
                    _ => None,
                }
            }
            MulExp::UnaryExp(exp) => exp.eval(scopes),
        }
    }
}

impl ExpEval for UnaryExp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            UnaryExp::Call(_) => None, // TODO: const fn
            UnaryExp::PrimaryExp(primary_exp) => primary_exp.eval(scopes),
            UnaryExp::UnaryOp(op, exp) => exp.eval(scopes).map(|exp| match op {
                UnaryOp::Add => exp,
                UnaryOp::Sub => -exp,
                UnaryOp::Not => (exp == 0) as i32,
            }),
        }
    }
}

impl ExpEval for PrimaryExp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            PrimaryExp::LVal(lval) => lval.eval(scopes),
            PrimaryExp::Number(number) => Some(*number),
            PrimaryExp::Paren(exp) => exp.eval(scopes),
        }
    }
}

impl ExpEval for LVal {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        let value = scopes.lookup_kv(&self.ident);
        match value {
            Some(ExpValue::Integer(integer)) => Some(integer),
            _ => None,
        }
    }
}

impl ExpEval for Exp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            Exp::LOrExp(lor_exp) => lor_exp.eval(scopes),
        }
    }
}

impl ExpEval for LOrExp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            LOrExp::LAndExp(land_exp) => land_exp.eval(scopes),
            LOrExp::Or(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some((lhs != 0 || rhs != 0) as i32),
                    _ => None,
                }
            }
        }
    }
}

impl ExpEval for LAndExp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            LAndExp::And(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some((lhs != 0 && rhs != 0) as i32),
                    _ => None,
                }
            }
            LAndExp::EqExp(eq_exp) => eq_exp.eval(scopes),
        }
    }
}

impl ExpEval for EqExp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            EqExp::Eq(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some((lhs == rhs) as i32),
                    _ => None,
                }
            }
            EqExp::Neq(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some((lhs != rhs) as i32),
                    _ => None,
                }
            }
            EqExp::RelExp(rel_exp) => rel_exp.eval(scopes),
        }
    }
}

impl ExpEval for RelExp {
    fn eval(&self, scopes: &mut Scopes) -> Option<i32> {
        match self {
            RelExp::AddExp(add_exp) => add_exp.eval(scopes),
            RelExp::Ge(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some((lhs >= rhs) as i32),
                    _ => None,
                }
            }
            RelExp::Gt(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some((lhs > rhs) as i32),
                    _ => None,
                }
            }
            RelExp::Le(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some((lhs <= rhs) as i32),
                    _ => None,
                }
            }
            RelExp::Lt(lhs, rhs) => {
                let lhs = lhs.eval(scopes);
                let rhs = rhs.eval(scopes);
                match (lhs, rhs) {
                    (Some(lhs), Some(rhs)) => Some((lhs < rhs) as i32),
                    _ => None,
                }
            }
        }
    }
}
