use koopa::ir::builder_traits::*;
use koopa::ir::*;

use crate::ast::{
    AddExp, AssignStmt, BType, BlockItem, BlockStmt, BreakStmt, CompUnit, Cond, ConstDecl,
    ConstDef, ConstInitVal, ContinueStmt, Decl, DeclOrFuncDef, EqExp, Exp, ExpStmt, FuncDef,
    FunctionCall, IfStmt, InitVal, LAndExp, LOrExp, LVal, MulExp, PrimaryExp, RelExp, ReturnStmt,
    Stmt, UnaryExp, UnaryOp, VarDecl, VarDef, WhileStmt,
};
use crate::eval::{ExpEval, ExpFunction, ExpValue};
use crate::scopes::{IRFunction, Scope, Scopes, WhileInfo};
use crate::{Error, Result};

pub trait IRGenerate {
    type Output;
    type AuxType;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output>;
}

pub fn declare_sysy_lib_functions(program: &mut Program, scopes: &mut Scopes) {
    /*
    decl @getint(): i32
    decl @getch(): i32
    decl @getarray(*i32): i32
    decl @putint(i32)
    decl @putch(i32)
    decl @putarray(i32, *i32)
    decl @starttime()
    decl @stoptime()
    */

    {
        // decl @getint(): i32
        let getint_funcdata = FunctionData::new_decl("@getint".into(), vec![], Type::get_i32());
        let getint_func = program.new_func(getint_funcdata);
        scopes.insert_kv(
            "getint".into(),
            ExpValue::Function(ExpFunction::new("@getint".into(), getint_func)),
        );
    }

    {
        // decl @getch(): i32
        let getch_funcdata = FunctionData::new_decl("@getch".into(), vec![], Type::get_i32());
        let getch_func = program.new_func(getch_funcdata);
        scopes.insert_kv(
            "getch".into(),
            ExpValue::Function(ExpFunction::new("@getch".into(), getch_func)),
        );
    }

    {
        // decl @getarray(*i32): i32
        let getarray_funcdata = FunctionData::new_decl(
            "@getarray".into(),
            vec![Type::get_pointer(Type::get_i32())],
            Type::get_i32(),
        );
        let getarray_func = program.new_func(getarray_funcdata);
        scopes.insert_kv(
            "getarray".into(),
            ExpValue::Function(ExpFunction::new("@getarray".into(), getarray_func)),
        );
    }

    {
        // decl @putint(i32)
        let putint_funcdata =
            FunctionData::new_decl("@putint".into(), vec![Type::get_i32()], Type::get_unit());
        let putint_func = program.new_func(putint_funcdata);
        scopes.insert_kv(
            "putint".into(),
            ExpValue::Function(ExpFunction::new("@putint".into(), putint_func)),
        );
    }

    {
        // decl @putch(i32)
        let putch_funcdata =
            FunctionData::new("@putch".into(), vec![Type::get_i32()], Type::get_unit());
        let putch_func = program.new_func(putch_funcdata);
        scopes.insert_kv(
            "putch".into(),
            ExpValue::Function(ExpFunction::new("@putch".into(), putch_func)),
        );
    }

    {
        // decl @putarray(i32, *i32)
        let putarray_funcdata = FunctionData::new(
            "@putarray".into(),
            vec![Type::get_i32(), Type::get_pointer(Type::get_i32())],
            Type::get_unit(),
        );
        let putarray_func = program.new_func(putarray_funcdata);
        scopes.insert_kv(
            "putarray".into(),
            ExpValue::Function(ExpFunction::new("@putarray".into(), putarray_func)),
        );
    }

    {
        // decl @starttime()
        let starttime_funcdata = FunctionData::new("@starttime".into(), vec![], Type::get_unit());
        let starttime_func = program.new_func(starttime_funcdata);
        scopes.insert_kv(
            "starttime".into(),
            ExpValue::Function(ExpFunction::new("@starttime".into(), starttime_func)),
        );
    }
    {
        // decl @stoptime()
        let stoptime_funcdata = FunctionData::new("@stoptime".into(), vec![], Type::get_unit());
        let stoptime_func = program.new_func(stoptime_funcdata);
        scopes.insert_kv(
            "stoptime".into(),
            ExpValue::Function(ExpFunction::new("@stoptime".into(), stoptime_func)),
        );
    }
}

impl IRGenerate for CompUnit {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        if let Some(ref prev_unit) = self.prev_unit {
            prev_unit.generate(program, scopes, ())?;
        }

        match &self.cur_unit {
            DeclOrFuncDef::FuncDef(func_def) => {
                func_def.generate(program, scopes, ())?;
            }
            DeclOrFuncDef::Decl(decl) => decl.generate(
                program,
                scopes,
                StmtDeclAux {
                    on_enter_renaming: true,
                    generate_next_basic_block: true,
                },
            )?,
        }

        Ok(())
    }
}

impl IRGenerate for Decl {
    type AuxType = StmtDeclAux;
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        // rename current basic block
        if !scopes.is_global_context() && aux.on_enter_renaming {
            let cur_func_mut = scopes.current_function_mut();
            cur_func_mut.rename_cur_block(
                program,
                Some(format!("%{}_decl", &cur_func_mut.name()[1..])),
            );
        }

        match self {
            Decl::ConstDecl(const_decl) => const_decl.generate(program, scopes, ())?,
            Decl::VarDecl(var_decl) => var_decl.generate(program, scopes, ())?,
        }

        if !scopes.is_global_context() && aux.generate_next_basic_block {
            let cur_func_mut = scopes.current_function_mut();

            // generate new basic block
            let next = cur_func_mut.new_basic_block(program, None);

            // jump to next basic block
            let cur_func_mut = scopes.current_function_mut();
            let jump = cur_func_mut.new_value(program).jump(next);
            cur_func_mut.push_inst(program, jump);

            // push the new basic block
            cur_func_mut.push_basic_block(program, next);
        }

        Ok(())
    }
}

impl IRGenerate for ConstDecl {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        for const_def in self.const_defs.iter() {
            const_def.generate(program, scopes, ConstDefAux { btype: self.btype })?;
        }
        Ok(())
    }
}

#[derive(Debug, Clone, Copy)]
pub struct ConstDefAux {
    btype: BType,
}

impl IRGenerate for ConstDef {
    type AuxType = ConstDefAux;
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        // Do not consider array for now

        let const_init_val = self.init_val.eval(scopes);

        if scopes.is_global_context() {
            match (aux.btype, const_init_val) {
                (BType::Int, Some(i)) => {
                    // add a global allocation
                    let const_init = if i == 0 {
                        program.new_value().zero_init(Type::get_i32())
                    } else {
                        program.new_value().integer(i)
                    };
                    let const_init_value = program.new_value().global_alloc(const_init);
                    program.set_value_name(const_init_value, Some(format!("@{}", self.ident)));
                    scopes.insert_kv(self.ident.clone(), ExpValue::InitVal(const_init_value));
                }
                (BType::Int, None) => {
                    // add a global allocation
                    let const_init = program.new_value().zero_init(Type::get_i32());
                    let const_init_value = program.new_value().global_alloc(const_init);
                    program.set_value_name(const_init_value, Some(format!("@{}", self.ident)));
                    scopes.insert_kv(self.ident.clone(), ExpValue::InitVal(const_init_value));
                }
                _ => panic!(),
            }
        }

        match (aux.btype, const_init_val) {
            (BType::Int, Some(i)) => {
                let const_init_val = ExpValue::Integer(i);
                scopes.insert_kv(self.ident.clone(), const_init_val);
            }
            _ => panic!(),
        }

        Ok(())
    }
}

impl IRGenerate for VarDecl {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        for var_def in self.var_defs.iter() {
            var_def.generate(program, scopes, VarDefAux { btype: self.btype })?;
        }
        Ok(())
    }
}

#[derive(Debug, Clone, Copy)]
pub struct VarDefAux {
    btype: BType,
}

impl IRGenerate for VarDef {
    type AuxType = VarDefAux;
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        // Do not consider array for now

        // TODO: redefinition in the same scope should err

        // handle global
        if scopes.is_global_context() {
            match self {
                VarDef::Init(ident, dims, init_val) => {
                    // Global definition with initial value
                    // And the value must be evaluated statically
                    let Some(i) = init_val.eval(scopes) else {
                        return Err(Error::IR(
                            "VarDef".into(),
                            "Global variable definition should be evaluated statically".into(),
                        ));
                    };
                    // For now do not consider about array, so zero_init i32
                    let init = if i == 0 {
                        program.new_value().zero_init(Type::get_i32())
                    } else {
                        program.new_value().integer(i)
                    };
                    let init_value = program.new_value().global_alloc(init);
                    program.set_value_name(init_value, Some(format!("@{}", ident)));
                    scopes.insert_kv(ident.clone(), ExpValue::InitVal(init_value));
                }
                VarDef::Uninit(ident, dims) => {
                    // Gloabl declaration, zero init
                    // For now do not consider about array, so zero_init i32
                    let zinit = program.new_value().zero_init(Type::get_i32());
                    let zinit_value = program.new_value().global_alloc(zinit);
                    program.set_value_name(zinit_value, Some(format!("@{}", ident)));
                    scopes.insert_kv(ident.clone(), ExpValue::InitVal(zinit_value));
                }
            }
            return Ok(());
        }

        let (ident, alloc) = match self {
            VarDef::Init(ident, dims, init_val) => {
                // Definition
                // generate init value
                let init_val = init_val.generate(program, scopes, ())?;
                let cur_func_mut = scopes.current_function_mut();
                // generate an alloc instruction in data flow graph
                let alloc = {
                    let alloc = cur_func_mut.new_value(program).alloc(aux.btype.into());
                    program
                        .func_mut(cur_func_mut.func())
                        .dfg_mut()
                        .set_value_name(alloc, Some(format!("%{}", ident)));
                    // push instruction into basic block
                    cur_func_mut.push_inst_to(program, cur_func_mut.entry_block(), alloc);
                    alloc
                };
                // generate a store instruction (need constant propagation here)
                let store_inst = cur_func_mut.new_value(program).store(init_val, alloc);
                cur_func_mut.push_inst(program, store_inst);
                (ident, alloc)
            }
            VarDef::Uninit(ident, dims) => {
                // Declaration
                // NOTE: alloc without name set
                // NOTE: do not generate store instruction
                let cur_func_mut = scopes.current_function_mut();
                let alloc = {
                    let alloc = cur_func_mut.new_value(program).alloc(aux.btype.into());
                    // push instruction into basic block
                    cur_func_mut.push_inst_to(program, cur_func_mut.entry_block(), alloc);
                    alloc
                };
                (ident, alloc)
            }
        };

        scopes.insert_kv(ident.into(), ExpValue::InitVal(alloc));
        Ok(())
    }
}

impl IRGenerate for InitVal {
    type AuxType = ();
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match self {
            InitVal::Exp(exp) => exp.generate(
                program,
                scopes,
                ExpAux {
                    on_assignment_left: false,
                },
            ),
            InitVal::Array(init_vals) => todo!(),
        }
    }
}

impl IRGenerate for FuncDef {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        // process function parameters.
        let params: Vec<(Option<String>, Type)> = self
            .params
            .0
            .iter()
            .map(|ffp| (Some(format!("@{}", ffp.ident)), ffp.btype.into()))
            .collect();

        let func_name = format!("@{}", self.ident);

        let mut data =
            FunctionData::with_param_names(func_name.clone(), params, self.func_type.into());
        let funcdata_params = data.params().to_owned();

        let data_mut = data.dfg_mut();
        let entry = data_mut
            .new_bb()
            .basic_block(Some(format!("%{}_entry", self.ident)));
        let exit = data_mut
            .new_bb()
            .basic_block(Some(format!("%{}_exit", self.ident)));
        let cur = data_mut.new_bb().basic_block(None);

        // generate return value
        let ret_val = match self.func_type {
            BType::Int => {
                let alloc = data_mut.new_value().alloc(Type::get_i32());
                data_mut.set_value_name(alloc, Some("%ret".into()));
                Some(alloc)
            }
            BType::Void => None,
        };

        let func = program.new_func(data);

        let mut ir_func = IRFunction::new(func_name.clone(), func, entry, exit, ret_val);
        ir_func.push_basic_block(program, entry);
        if let Some(ret_val) = ir_func.ret_val() {
            ir_func.push_inst(program, ret_val);
        }

        // record the function into current scope
        scopes.insert_kv(
            self.ident.clone(),
            ExpValue::Function(ExpFunction::new(func_name.clone(), func)),
        );

        // Set current function scope and generate all the basic blocks in this function.
        scopes.push_scope(Scope::new(format!("FuncDef_{}", func_name)));

        // generate parameters
        self.params
            .0
            .iter()
            .zip(funcdata_params)
            .for_each(|(ffp, value)| {
                let param_name = format!("%{}", ffp.ident);
                let param_type = program.func(func).dfg().value(value).ty().clone();

                let alloc_param = ir_func.new_value(program).alloc(param_type);
                ir_func
                    .func_data_mut(program)
                    .dfg_mut()
                    .set_value_name(alloc_param, Some(param_name));
                ir_func.push_inst(program, alloc_param);

                let store_param = ir_func.new_value(program).store(value, alloc_param);
                ir_func.push_inst(program, store_param);

                // add the parameter into current scope
                scopes.insert_kv(ffp.ident.clone(), ExpValue::InitVal(alloc_param));
            });

        ir_func.push_basic_block(program, cur);

        scopes.set_current_function(ir_func);
        let block_items_nums = self.block.items.len();
        for (i, block_item) in self.block.items.iter().enumerate() {
            block_item.generate(
                program,
                scopes,
                BlockItemAux {
                    is_first_block_item: i == 0,
                    is_last_block_item: i == (block_items_nums - 1),
                    inherited_stmtdecl_aux: StmtDeclAux {
                        on_enter_renaming: true,
                        generate_next_basic_block: true,
                    },
                },
            )?;
        }
        scopes.pop_scope();

        let mut ir_func = scopes.unset_current_function().unwrap();

        // entry -> second_bb
        let jump = ir_func.new_value(program).jump(ir_func.second_block());
        ir_func.push_inst_to(program, ir_func.entry_block(), jump);
        // cur -> exit
        ir_func.rename_cur_block(
            program,
            Some(format!("%{}_dummy_exit", &ir_func.name()[1..])),
        );
        let jump_to_exit = ir_func.new_value(program).jump(ir_func.exit_block());
        ir_func.push_inst(program, jump_to_exit);
        ir_func.push_basic_block(program, exit);
        let value = ir_func.ret_val().map(|alloc| {
            let value = ir_func.new_value(program).load(alloc);
            ir_func.push_inst(program, value);
            value
        });
        let ret = ir_func.new_value(program).ret(value);
        ir_func.push_inst(program, ret);
        Ok(())
    }
}

#[derive(Debug, Clone)]
pub struct BlockItemAux {
    is_first_block_item: bool,
    is_last_block_item: bool,

    inherited_stmtdecl_aux: StmtDeclAux,
}

/// Block item, corresponding to `basic block` in koopa IR.
impl IRGenerate for BlockItem {
    type AuxType = BlockItemAux;
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        // within a list of block items
        // the first block item should follow the order passed from parents whether to rename previous bb
        // the last block item should follow the order passed from parents whether to generate nexxt bb

        let on_enter_renaming = if aux.is_first_block_item {
            aux.inherited_stmtdecl_aux.on_enter_renaming
        } else {
            true // in the middle of the block item list, rename the previous one
        };

        let generate_next_basic_block = if aux.is_last_block_item {
            aux.inherited_stmtdecl_aux.generate_next_basic_block
        } else {
            true // in the middle of the block item list, always generate the next bb
        };

        match self {
            BlockItem::Decl(decl) => {
                decl.generate(
                    program,
                    scopes,
                    StmtDeclAux {
                        on_enter_renaming,
                        generate_next_basic_block,
                    },
                )?;
            }
            BlockItem::Stmt(stmt) => stmt.generate(
                program,
                scopes,
                StmtDeclAux {
                    on_enter_renaming,
                    generate_next_basic_block,
                },
            )?,
        }

        Ok(())
    }
}

#[derive(Debug, Clone)]
pub struct StmtDeclAux {
    on_enter_renaming: bool,
    generate_next_basic_block: bool,
}

impl IRGenerate for Stmt {
    type AuxType = StmtDeclAux;
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        // rename the current block
        if aux.on_enter_renaming {
            let cur_func_name = &scopes.current_function().name()[1..];
            let rename = if aux.on_enter_renaming {
                Some(match self {
                    Stmt::Assign(_) => format!("%{}_assignstmt", cur_func_name),
                    Stmt::Exp(_) => format!("%{}_expstmt", cur_func_name),
                    Stmt::Block(_) => format!("%{}_blockstmt", cur_func_name),
                    Stmt::If(_) => format!("%{}_ifstmt", cur_func_name),
                    Stmt::While(_) => format!("%{}_whilestmt", cur_func_name),
                    Stmt::Break(_) => format!("%{}_breakstmt", cur_func_name),
                    Stmt::Continue(_) => format!("%{}_continuestmt", cur_func_name),
                    Stmt::Return(_) => format!("%{}_returnstmt", cur_func_name),
                })
            } else {
                None
            };

            scopes
                .current_function_mut()
                .rename_cur_block(program, rename);
        }

        match self {
            Stmt::Assign(assign_stmt) => assign_stmt.generate(program, scopes, ())?,
            Stmt::Exp(exp_stmt) => exp_stmt.generate(program, scopes, ())?,
            Stmt::Block(block_stmt) => block_stmt.generate(program, scopes, aux.clone())?,
            Stmt::If(if_stmt) => if_stmt.generate(program, scopes, ())?,
            Stmt::While(while_stmt) => while_stmt.generate(program, scopes, ())?,
            Stmt::Break(break_stmt) => break_stmt.generate(program, scopes, ())?,
            Stmt::Continue(continue_stmt) => continue_stmt.generate(program, scopes, ())?,
            Stmt::Return(return_stmt) => return_stmt.generate(program, scopes, ())?,
        }

        let cur_func_mut = scopes.current_function_mut();
        let next = if aux.generate_next_basic_block && !matches!(self, Stmt::Block(_)) {
            // generate new basic block
            let next = cur_func_mut.new_basic_block(program, None);
            Some(next)
        } else {
            None
        };

        let jump_target = match self {
            Stmt::Return(_) => Some(cur_func_mut.exit_block()),
            _ => next,
        };

        if let Some(jump_target) = jump_target {
            // jump to next basic block
            let jump = cur_func_mut.new_value(program).jump(jump_target);
            cur_func_mut.push_inst(program, jump);
        }

        if let Some(next) = next {
            // push the new basic block
            cur_func_mut.push_basic_block(program, next);
        }

        Ok(())
    }
}

impl IRGenerate for WhileStmt {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        let cur_func_name = &scopes.current_function().name()[1..];
        let entry_bb_name = format!("%{}_while_entry_bb", cur_func_name);
        let body_bb_name = format!("%{}_while_body_bb", cur_func_name);
        let end_bb_name = format!("%{}_while_end_bb", cur_func_name);

        // generate entry, body and end basic block
        let cur_func_mut = scopes.current_function_mut();

        let entry_bb = cur_func_mut.new_basic_block(program, Some(&entry_bb_name));
        let body_bb = cur_func_mut.new_basic_block(program, Some(&body_bb_name));
        let end_bb = cur_func_mut.new_basic_block(program, Some(&end_bb_name));

        // jump into entry bb
        let jump_into_entry_inst = cur_func_mut.new_value(program).jump(entry_bb);
        cur_func_mut.push_inst(program, jump_into_entry_inst);

        // generate entry bb
        cur_func_mut.push_basic_block(program, entry_bb);

        // calculate condition code
        let cond_val = self.cond.generate(
            program,
            scopes,
            ExpAux {
                on_assignment_left: false,
            },
        )?;

        let cur_func_mut = scopes.current_function_mut();
        // branch instruction
        let cond_branch_inst = cur_func_mut
            .new_value(program)
            .branch(cond_val, body_bb, end_bb);
        cur_func_mut.push_inst(program, cond_branch_inst);

        // push a layer of while info
        scopes.push_while_info(WhileInfo::new(entry_bb, body_bb, end_bb));
        let cur_func_mut = scopes.current_function_mut();
        // generate body basic block
        cur_func_mut.push_basic_block(program, body_bb);
        self.body_stmt.generate(
            program,
            scopes,
            StmtDeclAux {
                on_enter_renaming: false,         // do not rename because already named
                generate_next_basic_block: false, // do not generate next basic block because manually do
            },
        )?;
        scopes.pop_while_info();

        let cur_func_mut = scopes.current_function_mut();
        let body_last_inst = cur_func_mut.inspect_last_inst(program);
        match body_last_inst {
            Some(body_last_inst) if matches!(body_last_inst.kind(), ValueKind::Jump(jump) if jump.target() == cur_func_mut.exit_block()) =>
                {}
            _ => {
                // generate jumping to entry block
                let jump_into_entry_inst = cur_func_mut.new_value(program).jump(entry_bb);
                cur_func_mut.push_inst(program, jump_into_entry_inst);
            }
        }

        // generate end basic block
        cur_func_mut.push_basic_block(program, end_bb);

        Ok(())
    }
}

impl IRGenerate for BreakStmt {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        let Some(cur_while_info) = scopes.cur_while_info() else {
            return Err(Error::IR(
                "BreakStmt".into(),
                "Not inside a while body".into(),
            ));
        };

        // jump to end block
        let cur_func_mut = scopes.current_function_mut();
        let jump_inst = cur_func_mut
            .new_value(program)
            .jump(cur_while_info.end_bb());
        cur_func_mut.push_inst(program, jump_inst);

        let after_break_bb = cur_func_mut.new_basic_block(
            program,
            Some(&format!("%{}_after_break", &cur_func_mut.name()[1..])),
        );
        cur_func_mut.push_basic_block(program, after_break_bb);

        Ok(())
    }
}

impl IRGenerate for ContinueStmt {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        let Some(cur_while_info) = scopes.cur_while_info() else {
            return Err(Error::IR(
                "ContinueStmt".into(),
                "Not inside a while body".into(),
            ));
        };

        // jump to end block
        let cur_func_mut = scopes.current_function_mut();
        let jump_inst = cur_func_mut
            .new_value(program)
            .jump(cur_while_info.entry_bb());
        cur_func_mut.push_inst(program, jump_inst);

        let after_break_bb = cur_func_mut.new_basic_block(
            program,
            Some(&format!("%{}_after_continue", &cur_func_mut.name()[1..])),
        );
        cur_func_mut.push_basic_block(program, after_break_bb);

        Ok(())
    }
}

impl IRGenerate for IfStmt {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        let cur_func_name = &scopes.current_function().name()[1..];
        let thenbb_name = format!("%{}_if_then_bb", cur_func_name);
        let elsebb_name = format!("%{}_if_else_bb", cur_func_name);
        let endbb_name = format!("%{}_if_end_bb", cur_func_name);

        // First calculate condition code
        let cond_val = self.cond.generate(
            program,
            scopes,
            ExpAux {
                on_assignment_left: false,
            },
        )?;

        let cur_func_mut = scopes.current_function_mut();

        // generate then, else and end basic blocks
        let then_bb = cur_func_mut.new_basic_block(program, Some(&thenbb_name));
        let else_bb = cur_func_mut.new_basic_block(program, Some(&elsebb_name));
        let end_bb = cur_func_mut.new_basic_block(program, Some(&endbb_name));

        // branch instruction
        let branch_inst = cur_func_mut
            .new_value(program)
            // .branch(cond_load_inst, then_bb, else_bb);
            .branch(cond_val, then_bb, else_bb);
        cur_func_mut.push_inst(program, branch_inst);

        // generate then basic block
        cur_func_mut.push_basic_block(program, then_bb);
        self.then_stmt.generate(
            program,
            scopes,
            StmtDeclAux {
                on_enter_renaming: false,         // do not rename because already named
                generate_next_basic_block: false, // do not generate next basic block because manually do
            },
        )?;

        // generate jumping to end block.
        // however, if `then` statement ends with a return statement
        // which means it has already jumped to exit block
        // then should not generate jumping to end inst.

        // inspect last inst
        let cur_func_mut = scopes.current_function_mut();
        let then_bb_last_inst = cur_func_mut.inspect_last_inst(program);
        match then_bb_last_inst {
            Some(then_bb_last_inst) if matches!(then_bb_last_inst.kind(), ValueKind::Jump(jump) if jump.target() == cur_func_mut.exit_block()) =>
                {}
            _ => {
                let jump_from_then_to_end = cur_func_mut.new_value(program).jump(end_bb);
                cur_func_mut.push_inst(program, jump_from_then_to_end);
            }
        }

        // generate else basic block
        cur_func_mut.push_basic_block(program, else_bb);
        match &self.else_stmt {
            Some(else_stmt) => else_stmt.generate(
                program,
                scopes,
                StmtDeclAux {
                    on_enter_renaming: false,         // do not rename because already named
                    generate_next_basic_block: false, // do not generate next basic block because manually do
                },
            )?,
            None => (),
        }

        // inspect last inst
        let cur_func_mut = scopes.current_function_mut();
        let then_bb_last_inst = cur_func_mut.inspect_last_inst(program);
        match then_bb_last_inst {
            Some(then_bb_last_inst) if matches!(then_bb_last_inst.kind(), ValueKind::Jump(jump) if jump.target() == cur_func_mut.exit_block()) =>
                {}
            _ => {
                let jump_from_else_to_end = cur_func_mut.new_value(program).jump(end_bb);
                cur_func_mut.push_inst(program, jump_from_else_to_end);
            }
        }

        // push end block
        cur_func_mut.push_basic_block(program, end_bb);

        Ok(())
    }
}

impl IRGenerate for Cond {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        let val = match self {
            Cond::LOrExp(lor_exp) => lor_exp.generate(program, scopes, aux)?,
        };
        Ok(val)
    }
}

impl IRGenerate for ExpStmt {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match &self.exp {
            Some(exp) => {
                // the value returned is discarded
                let _ = exp.generate(
                    program,
                    scopes,
                    ExpAux {
                        on_assignment_left: false,
                    },
                );
            }
            None => {}
        }

        Ok(())
    }
}

impl IRGenerate for BlockStmt {
    type AuxType = StmtDeclAux;
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        let cur_func_name = &scopes.current_function().name()[1..];
        let scope_name = format!("Block_{}", cur_func_name);

        // Set current function scope and generate all the basic blocks in this function.
        scopes.push_scope(Scope::new(scope_name));

        let block_items_nums = self.block.items.len();
        for (i, block_item) in self.block.items.iter().enumerate() {
            block_item.generate(
                program,
                scopes,
                BlockItemAux {
                    is_first_block_item: i == 0,
                    is_last_block_item: i == (block_items_nums - 1),
                    inherited_stmtdecl_aux: aux.clone(),
                },
            )?;
        }
        scopes.pop_scope();

        // let cur_func_mut = scopes.current_function_mut();

        // // generate new basic block
        // let next = cur_func_mut.new_basic_block(program, None);

        // // jump to next basic block
        // let cur_func_mut = scopes.current_function_mut();
        // let jump = cur_func_mut.new_value(program).jump(next);
        // cur_func_mut.push_inst(program, jump);

        // // push the new basic block
        // cur_func_mut.push_basic_block(program, next);

        Ok(())
    }
}

impl IRGenerate for AssignStmt {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        let exp = self.exp.generate(
            program,
            scopes,
            ExpAux {
                on_assignment_left: false,
            },
        )?;
        let lval = self.lval.generate(
            program,
            scopes,
            ExpAux {
                on_assignment_left: true,
            },
        )?;

        // generate store
        let cur_func_mut = scopes.current_function_mut();
        let store_inst = cur_func_mut.new_value(program).store(exp, lval);
        cur_func_mut.push_inst(program, store_inst);

        Ok(())
    }
}

impl IRGenerate for ReturnStmt {
    type AuxType = ();
    type Output = ();

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        // evaluate the Exp
        match &self.exp {
            Some(ret_exp) => {
                if let Some(ret_val) = scopes.current_function().ret_val() {
                    // generate store
                    let value = ret_exp.generate(
                        program,
                        scopes,
                        ExpAux {
                            on_assignment_left: false,
                        },
                    )?;
                    let ir_func = scopes.current_function();
                    let store = ir_func.new_value(program).store(value, ret_val);
                    ir_func.push_inst(program, store);
                } else {
                    // have ret_exp but function type do not hint a return value
                    return Err(Error::MalformedReturn);
                }
            }
            None => {
                // returning void, do nothing
            }
        }

        Ok(())
    }
}

#[derive(Debug, Clone)]
pub struct ExpAux {
    on_assignment_left: bool,
}

impl IRGenerate for LVal {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        // Try to find value in the scope
        let Some(init_val) = scopes.lookup_kv(&self.ident) else {
            return Err(Error::IR(
                "LVal".into(),
                format!("Fail to find init value for LVal {} in scopes", self.ident),
            ));
        };

        let cur_func_mut = scopes.current_function_mut();

        // There should be a alloc inst
        let the_val = match init_val {
            ExpValue::InitVal(val) => {
                if aux.on_assignment_left {
                    val // if lval on the left side of assignment
                } else {
                    // first load the value
                    let load_inst = cur_func_mut.new_value(program).load(val);
                    cur_func_mut.push_inst(program, load_inst);

                    let load_inst_value_data =
                        program.func(cur_func_mut.func()).dfg().value(load_inst);
                    dbg!(load_inst_value_data);

                    load_inst // if lval on the right side of assignment
                }
            }
            ExpValue::Integer(integer) => {
                // come from const eval?
                cur_func_mut.new_value(program).integer(integer)
            }
            _ => {
                return Err(Error::IR(
                    "LVal".into(),
                    format!("Fail to find init value for LVal {} in scopes", self.ident),
                ))
            }
        };

        Ok(the_val)
    }
}

impl IRGenerate for Exp {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        let Self::LOrExp(ref lor_exp) = self;
        lor_exp.generate(program, scopes, aux)
    }
}

impl IRGenerate for PrimaryExp {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match self {
            PrimaryExp::LVal(lval) => lval.generate(
                program,
                scopes,
                ExpAux {
                    on_assignment_left: false,
                },
            ),
            PrimaryExp::Number(number) => {
                let n = scopes
                    .current_function()
                    .new_value(program)
                    .integer(*number);
                Ok(n)
            }
            PrimaryExp::Paren(exp) => exp.generate(program, scopes, aux),
        }
    }
}

impl IRGenerate for UnaryExp {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match self {
            UnaryExp::Call(call) => call.generate(program, scopes, aux),
            UnaryExp::UnaryOp(unary_op, unary_exp) => {
                let unary_exp = unary_exp.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();

                // Since there are only binary op in Koopa IR...
                let zero = cur_func_mut.new_value(program).integer(0);
                let unary_inst = match unary_op {
                    // Not operation
                    UnaryOp::Not => {
                        cur_func_mut
                            .new_value(program)
                            .binary(BinaryOp::Eq, unary_exp, zero)
                    }
                    // Negative operation
                    UnaryOp::Sub => {
                        cur_func_mut
                            .new_value(program)
                            .binary(BinaryOp::Sub, zero, unary_exp)
                    }
                    UnaryOp::Add => {
                        cur_func_mut
                            .new_value(program)
                            .binary(BinaryOp::Add, unary_exp, zero)
                    }
                };
                cur_func_mut.push_inst(program, unary_inst);
                Ok(unary_inst)
            }
            UnaryExp::PrimaryExp(primary_exp) => primary_exp.generate(program, scopes, aux),
        }
    }
}

impl IRGenerate for FunctionCall {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        // Look up in scopes to find the function
        let target_func = scopes.lookup_kv(&self.ident);
        let Some(ExpValue::Function(target_func)) = target_func else {
            return Err(Error::IR(
                "FunctionCall".into(),
                format!("Function {} not found", self.ident),
            ));
        };

        // TODO: Check return value

        // TODO: Check parameters

        // Calculate args
        let mut args = vec![];
        for param in self.params.0.iter() {
            let arg = param.generate(program, scopes, aux.clone())?;
            args.push(arg);
        }

        let cur_func_mut = scopes.current_function_mut();
        let call_inst = cur_func_mut
            .new_value(program)
            .call(target_func.func(), args);
        cur_func_mut.push_inst(program, call_inst);

        Ok(call_inst)
    }
}

impl IRGenerate for MulExp {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match self {
            MulExp::UnaryExp(unary_exp) => unary_exp.generate(program, scopes, aux),
            MulExp::Mul(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let add_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Mul, lhs, rhs);
                cur_func_mut.push_inst(program, add_inst);
                Ok(add_inst)
            }
            MulExp::Div(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let add_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Div, lhs, rhs);
                cur_func_mut.push_inst(program, add_inst);
                Ok(add_inst)
            }
            MulExp::Mod(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let add_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Mod, lhs, rhs);
                cur_func_mut.push_inst(program, add_inst);
                Ok(add_inst)
            }
        }
    }
}

impl IRGenerate for AddExp {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match self {
            AddExp::Add(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();

                let lhs_value_data = program.func(cur_func_mut.func()).dfg().value(lhs);
                let rhs_value_data = program.func(cur_func_mut.func()).dfg().value(rhs);
                dbg!(lhs_value_data);
                dbg!(rhs_value_data);

                let add_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Add, lhs, rhs);
                cur_func_mut.push_inst(program, add_inst);
                Ok(add_inst)
            }
            AddExp::Sub(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let add_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Sub, lhs, rhs);
                cur_func_mut.push_inst(program, add_inst);
                Ok(add_inst)
            }
            AddExp::MulExp(mul_exp) => mul_exp.generate(program, scopes, aux),
        }
    }
}

impl IRGenerate for RelExp {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match self {
            RelExp::AddExp(add_exp) => add_exp.generate(program, scopes, aux),
            RelExp::Ge(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let ge_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Ge, lhs, rhs);
                cur_func_mut.push_inst(program, ge_inst);
                Ok(ge_inst)
            }
            RelExp::Gt(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let gt_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Gt, lhs, rhs);
                cur_func_mut.push_inst(program, gt_inst);
                Ok(gt_inst)
            }
            RelExp::Le(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let le_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Le, lhs, rhs);
                cur_func_mut.push_inst(program, le_inst);
                Ok(le_inst)
            }
            RelExp::Lt(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let lt_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Lt, lhs, rhs);
                cur_func_mut.push_inst(program, lt_inst);
                Ok(lt_inst)
            }
        }
    }
}

impl IRGenerate for EqExp {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match self {
            EqExp::Eq(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let eq_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::Eq, lhs, rhs);
                cur_func_mut.push_inst(program, eq_inst);
                Ok(eq_inst)
            }
            EqExp::Neq(lhs, rhs) => {
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let neq_inst = cur_func_mut
                    .new_value(program)
                    .binary(BinaryOp::NotEq, lhs, rhs);
                cur_func_mut.push_inst(program, neq_inst);
                Ok(neq_inst)
            }
            EqExp::RelExp(rel_exp) => rel_exp.generate(program, scopes, aux),
        }
    }
}

impl IRGenerate for LAndExp {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match self {
            LAndExp::And(lhs, rhs) => {
                /*
                int result = 1;
                if (lhs == 0) {

                } else {
                    result = rhs != 0;
                }
                */

                // allocate a result position
                let cur_func_mut = scopes.current_function_mut();
                let result = cur_func_mut.new_value(program).alloc(Type::get_i32());
                cur_func_mut.push_inst_to(program, cur_func_mut.entry_block(), result);

                // generate lhs
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let cur_func_mut = scopes.current_function_mut();
                let zero = cur_func_mut.new_value(program).integer(0);
                let comp_lhs_inst =
                    cur_func_mut
                        .new_value(program)
                        .binary(BinaryOp::NotEq, zero, lhs);
                cur_func_mut.push_inst(program, comp_lhs_inst);

                // store lhs comp into result
                let store_lhs_inst = cur_func_mut.new_value(program).store(comp_lhs_inst, result);
                cur_func_mut.push_inst(program, store_lhs_inst);

                // generate rhs and end bb
                let cur_func_name = &cur_func_mut.name()[1..];
                let rhs_bb_name = format!("%{}_andrhsbb", cur_func_name);
                let end_bb_name = format!("%{}_andendbb", cur_func_name);
                let rhs_bb = cur_func_mut.new_basic_block(program, Some(&rhs_bb_name));
                let end_bb = cur_func_mut.new_basic_block(program, Some(&end_bb_name));

                // generate branch inst
                let branch_inst =
                    cur_func_mut
                        .new_value(program)
                        .branch(comp_lhs_inst, rhs_bb, end_bb);
                cur_func_mut.push_inst(program, branch_inst);

                // push rhs bb and generate rhs
                cur_func_mut.push_basic_block(program, rhs_bb);

                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let comp_rhs_inst =
                    cur_func_mut
                        .new_value(program)
                        .binary(BinaryOp::NotEq, zero, rhs);
                cur_func_mut.push_inst(program, comp_rhs_inst);

                // store rhs comp into result
                let store_rhs_inst = cur_func_mut.new_value(program).store(comp_rhs_inst, result);
                cur_func_mut.push_inst(program, store_rhs_inst);

                // jump to end bb
                let jump_end_inst = cur_func_mut.new_value(program).jump(end_bb);
                cur_func_mut.push_inst(program, jump_end_inst);

                cur_func_mut.push_basic_block(program, end_bb);

                // load result as cond / val
                let load_inst = cur_func_mut.new_value(program).load(result);
                cur_func_mut.push_inst(program, load_inst);
                Ok(load_inst)

                // let and_inst = cur_func_mut.new_value(program).binary(
                //     BinaryOp::And,
                //     comp_lhs_inst,
                //     comp_rhs_inst,
                // );
                // cur_func_mut.push_inst(program, and_inst);
                // Ok(and_inst)
            }
            LAndExp::EqExp(eq_exp) => eq_exp.generate(program, scopes, aux),
        }
    }
}

impl IRGenerate for LOrExp {
    type AuxType = ExpAux;
    type Output = Value;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        match self {
            LOrExp::LAndExp(land_exp) => land_exp.generate(program, scopes, aux),
            LOrExp::Or(lhs, rhs) => {
                /*
                int result = 1;
                if (lhs == 0) {
                    result = rhs != 0;
                }
                */

                // allocate a result position
                let cur_func_mut = scopes.current_function_mut();
                let result = cur_func_mut.new_value(program).alloc(Type::get_i32());
                cur_func_mut.push_inst_to(program, cur_func_mut.entry_block(), result);

                // generate lhs
                let lhs = lhs.generate(program, scopes, aux.clone())?;
                let cur_func_mut = scopes.current_function_mut();
                let zero = cur_func_mut.new_value(program).integer(0);
                let comp_lhs_inst =
                    cur_func_mut
                        .new_value(program)
                        .binary(BinaryOp::NotEq, zero, lhs);
                cur_func_mut.push_inst(program, comp_lhs_inst);

                // store lhs comp into result
                let store_lhs_inst = cur_func_mut.new_value(program).store(comp_lhs_inst, result);
                cur_func_mut.push_inst(program, store_lhs_inst);

                // generate rhs and end bb
                let cur_func_name = &cur_func_mut.name()[1..];
                let rhs_bb_name = format!("%{}_orrhsbb", cur_func_name);
                let end_bb_name = format!("%{}_orendbb", cur_func_name);
                let rhs_bb = cur_func_mut.new_basic_block(program, Some(&rhs_bb_name));
                let end_bb = cur_func_mut.new_basic_block(program, Some(&end_bb_name));

                // generate branch inst
                let branch_inst =
                    cur_func_mut
                        .new_value(program)
                        .branch(comp_lhs_inst, end_bb, rhs_bb);
                cur_func_mut.push_inst(program, branch_inst);

                // push rhs bb and generate rhs
                cur_func_mut.push_basic_block(program, rhs_bb);
                let rhs = rhs.generate(program, scopes, aux)?;
                let cur_func_mut = scopes.current_function_mut();
                let comp_rhs_inst =
                    cur_func_mut
                        .new_value(program)
                        .binary(BinaryOp::NotEq, zero, rhs);
                cur_func_mut.push_inst(program, comp_rhs_inst);

                // store rhs comp into result
                let store_rhs_inst = cur_func_mut.new_value(program).store(comp_rhs_inst, result);
                cur_func_mut.push_inst(program, store_rhs_inst);

                // jump to end bb
                let jump_end_inst = cur_func_mut.new_value(program).jump(end_bb);
                cur_func_mut.push_inst(program, jump_end_inst);

                cur_func_mut.push_basic_block(program, end_bb);

                // load result as cond / val
                let load_inst = cur_func_mut.new_value(program).load(result);
                cur_func_mut.push_inst(program, load_inst);
                Ok(load_inst)

                // let or_inst = cur_func_mut.new_value(program).binary(
                //     BinaryOp::Or,
                //     comp_lhs_inst,
                //     comp_rhs_inst,
                // );
                // cur_func_mut.push_inst(program, or_inst);
                // Ok(or_inst)
            }
        }
    }
}

impl From<BType> for koopa::ir::Type {
    fn from(bt: BType) -> Self {
        match bt {
            BType::Int => koopa::ir::Type::get_i32(),
            BType::Void => koopa::ir::Type::get_unit(),
        }
    }
}

impl IRGenerate for BType {
    type AuxType = ();
    type Output = koopa::ir::Type;

    fn generate(
        &self,
        program: &mut Program,
        scopes: &mut Scopes,
        aux: Self::AuxType,
    ) -> Result<Self::Output> {
        let ret = match self {
            BType::Int => koopa::ir::Type::get_i32(),
            BType::Void => koopa::ir::Type::get_unit(),
        };

        Ok(ret)
    }
}
