use crate::koopa::declare_sysy_lib_functions;
use crate::koopa::IRGenerate;
use ::koopa::back::KoopaGenerator;
use asm_aux::{AsmProgram, FileEmitter, StdoutEmitter};
use cmdline::CommandLineArgs;
use lexer::Lexer;
use scopes::Scope;
use std::fs;

mod asm;
mod asm_aux;
mod ast;
mod cmdline;
mod eval;
mod koopa;
mod lexer;
mod scopes;
mod sysy; // generate_in_source_tree
mod token;

#[derive(Debug, thiserror::Error)]
pub enum Error {
    #[error("I/O Error: {0}")]
    IO(#[from] std::io::Error),
    #[error("Invalid variable: {0}")]
    InvalidVar(String),
    #[error("IR generation error when {0}: {1}")]
    IR(String, String),
    #[error("Asm generation error when {0}: {1}")]
    Asm(String, String),
    #[error("Function malformed return")]
    MalformedReturn,
}

pub type Result<T> = std::result::Result<T, Error>;

fn check_var(name: &str, var: Option<&String>) -> Result<()> {
    if var.is_none() || var.unwrap().is_empty() {
        return Err(Error::InvalidVar(format!("Empty variable {name}")));
    }

    Ok(())
}

fn main() -> Result<()> {
    let cmdline = CommandLineArgs::parse();

    let mut gen_riscv = false;
    let koopa_src_file_path;

    // Check the path
    if check_var("-koopa", cmdline.koopa()).is_err() {
        check_var("-riscv", cmdline.riscv())?;
        gen_riscv = true;
        koopa_src_file_path = cmdline.riscv().unwrap();
    } else {
        koopa_src_file_path = cmdline.koopa().unwrap();
    }
    check_var("-o", cmdline.output())?;

    let koopa_output_file_path = cmdline.output().unwrap();

    let source_code = fs::read_to_string(koopa_src_file_path)?;
    let lexer = Lexer::new(&source_code);
    let parser = sysy::CompUnitParser::new();
    let ast = parser.parse(lexer).unwrap();
    println!("{:#?}", ast);

    let mut program = ::koopa::ir::Program::new();
    let mut scopes = scopes::Scopes::new();
    scopes.push_scope(Scope::new("Global")); // Global scope
    declare_sysy_lib_functions(&mut program, &mut scopes); // sysy library functions

    ast.generate(&mut program, &mut scopes, ())?;

    if !gen_riscv {
        KoopaGenerator::from_path(koopa_output_file_path)?.generate_on(&program)?;
    } else {
        let asm_file = fs::OpenOptions::new()
            .create(true)
            .write(true)
            .truncate(true)
            .open(koopa_output_file_path)?;
        let emitter = FileEmitter::new(asm_file);
        // let emitter= StdoutEmitter::new();
        let mut asm_program = AsmProgram::new(&program);
        use asm::AsmGenerate;
        program.generate(&emitter, &mut asm_program, ())?;
    }

    Ok(())
}
