use std::collections::HashMap;

use koopa::ir::{
    builder::{BasicBlockBuilder, LocalBuilder, LocalInstBuilder},
    entities::ValueData,
    BasicBlock, Function, FunctionData, Program, Value,
};

use crate::eval::ExpValue;

#[derive(Debug)]
pub struct Scopes {
    inner: Vec<Scope>,
    current_function: Option<IRFunction>, // this is Copy
    while_infos: Vec<WhileInfo>,
}

#[derive(Debug, Clone)]
pub struct WhileInfo {
    entry_bb: BasicBlock,
    body_bb: BasicBlock,
    end_bb: BasicBlock,
}

impl WhileInfo {
    pub fn new(entry_bb: BasicBlock, body_bb: BasicBlock, end_bb: BasicBlock) -> Self {
        Self {
            entry_bb,
            body_bb,
            end_bb,
        }
    }

    pub fn entry_bb(&self) -> BasicBlock {
        self.entry_bb
    }

    pub fn body_bb(&self) -> BasicBlock {
        self.body_bb
    }

    pub fn end_bb(&self) -> BasicBlock {
        self.end_bb
    }
}

impl Scopes {
    pub fn new() -> Self {
        Self {
            inner: vec![],
            current_function: None,
            while_infos: vec![],
        }
    }

    pub fn push_while_info(&mut self, info: WhileInfo) {
        self.while_infos.push(info);
    }

    pub fn pop_while_info(&mut self) {
        self.while_infos
            .pop()
            .expect("Fail to pop while info, check generator");
    }

    pub fn cur_while_info(&self) -> Option<WhileInfo> {
        self.while_infos.last().cloned()
    }

    pub fn push_scope(&mut self, scope: Scope) {
        self.inner.push(scope);
    }

    pub fn pop_scope(&mut self) -> Option<Scope> {
        self.inner.pop()
    }

    pub fn top_scope(&self) -> Option<&Scope> {
        self.inner.last()
    }

    pub fn is_global_context(&self) -> bool {
        self.inner.len() == 1
    }

    pub fn top_scope_mut(&mut self) -> Option<&mut Scope> {
        self.inner.last_mut()
    }

    pub fn set_current_function(&mut self, func: IRFunction) {
        self.current_function = Some(func);
    }

    pub fn unset_current_function(&mut self) -> Option<IRFunction> {
        self.current_function.take()
    }

    pub fn current_function(&self) -> &IRFunction {
        self.current_function
            .as_ref()
            // .map(|(_, x)| program.func_mut(*x))
            .unwrap_or_else(|| panic!("Check generator!"))
    }

    pub fn current_function_mut(&mut self) -> &mut IRFunction {
        self.current_function
            .as_mut()
            // .map(|(_, x)| program.func_mut(*x))
            .unwrap_or_else(|| panic!("Check generator!"))
    }

    pub fn print_status(&self) {
        println!(
            "Current Function: {}",
            self.current_function
                .as_ref()
                .map(|func| &func.name)
                .unwrap_or(&"NULL".into())
        );
        self.inner.iter().enumerate().for_each(|(i, scope)| {
            scope.print_status(i);
        });
    }

    pub fn lookup_kv(&self, key: &String) -> Option<ExpValue> {
        self.inner
            .iter()
            .rev()
            .skip_while(|scope| scope.lookup_kv(key).is_none())
            .next()
            .and_then(|s| s.lookup_kv(key))
    }

    pub fn insert_kv(&mut self, key: String, value: ExpValue) {
        self.top_scope_mut().unwrap().insert_kv(key, value);
    }
}

#[derive(Debug)]
pub struct Scope {
    name: String,
    kvmap: HashMap<String, ExpValue>,
}

impl Scope {
    pub fn new<S: AsRef<str>>(name: S) -> Self {
        Self {
            name: name.as_ref().into(),
            kvmap: HashMap::new(),
        }
    }

    fn insert_kv(&mut self, key: String, value: ExpValue) {
        self.kvmap.insert(key, value);
    }

    fn lookup_kv(&self, key: &String) -> Option<ExpValue> {
        self.kvmap.get(key).cloned()
    }

    pub fn print_status(&self, i: usize) {
        println!("Scope {i} ================ {} ================", self.name);
        self.kvmap.iter().for_each(|(k, v)| {
            println!("     + {k:<20} : {v:#?}");
        });
    }
}

#[derive(Debug, Clone)]
pub struct IRFunction {
    name: String,
    func: Function,
    entry: BasicBlock,
    second: BasicBlock,
    exit: BasicBlock,
    cur: BasicBlock,
    ret_val: Option<Value>,
}

impl IRFunction {
    pub fn new(
        name: String,
        func: Function,
        entry: BasicBlock,
        exit: BasicBlock,
        ret_val: Option<Value>,
    ) -> Self {
        Self {
            name,
            func,
            entry,
            second: entry,
            exit,
            cur: entry,
            ret_val,
        }
    }

    pub fn func(&self) -> Function {
        self.func
    }

    pub fn func_data_mut<'a>(&'a self, program: &'a mut Program) -> &'a mut FunctionData {
        program.func_mut(self.func)
    }

    pub fn name(&self) -> &String {
        &self.name
    }

    pub fn entry_block(&self) -> BasicBlock {
        self.entry
    }

    pub fn second_block(&self) -> BasicBlock {
        self.second
    }

    pub fn exit_block(&self) -> BasicBlock {
        self.exit
    }

    pub fn ret_val(&self) -> Option<Value> {
        self.ret_val
    }

    pub fn push_basic_block(&mut self, program: &mut Program, bb: BasicBlock) {
        program
            .func_mut(self.func)
            .layout_mut()
            .bbs_mut()
            .push_key_back(bb)
            .unwrap();
        self.cur = bb;
        if self.second == self.entry {
            self.second = bb;
        }
    }

    pub fn push_inst_to(&self, program: &mut Program, basic_block: BasicBlock, inst: Value) {
        program
            .func_mut(self.func)
            .layout_mut()
            .bb_mut(basic_block)
            .insts_mut()
            .push_key_back(inst)
            .unwrap();
    }

    pub fn push_inst(&self, program: &mut Program, inst: Value) {
        self.push_inst_to(program, self.cur, inst);
    }

    pub fn inspect_last_inst(&mut self, program: &mut Program) -> Option<ValueData> {
        program
            .func_mut(self.func)
            .layout_mut()
            .bb_mut(self.cur)
            .insts()
            .back_key()
            .cloned()
            .and_then(|v| Some(program.func(self.func).dfg().value(v).clone()))
    }

    pub fn new_basic_block(&self, program: &mut Program, name: Option<&str>) -> BasicBlock {
        program
            .func_mut(self.func)
            .dfg_mut()
            .new_bb()
            .basic_block(name.map(|s| s.into()))
    }

    pub fn rename_cur_block(&self, program: &mut Program, name: Option<String>) {
        program
            .func_mut(self.func)
            .dfg_mut()
            .bb_mut(self.cur)
            .set_name(name);
    }

    pub fn new_value<'p>(&self, program: &'p mut Program) -> LocalBuilder<'p> {
        program.func_mut(self.func).dfg_mut().new_value()
    }
}
